package com.getsnappy.snappyj;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.net.URLEncoder;
import java.util.HashMap;
import java.util.Map;
import java.util.Vector;

import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.log4j.Logger;

/**
 * These beans act as the communication channel between javascript and java.
 * In javascript beans are manipulated and saved cookies.  Each bean has its 
 * own cookie.  These cookies must always be maintained to represent the 
 * current state, and are used in case of a reload.  All variables in all
 * beans can also be passed in through the request which overrides the cookie.
 * 
 * 
 * @author brian
 *
 */
public abstract class JavaScriptBean {
   private static Logger log = Logger.getLogger(JavaScriptBean.class);
   private boolean isDirty = false;
   
   /**
    *   // overide this function to define a bean name.  This function should be a return a static string and you should
    *   // implement a static version as well
    *   //
    *   // e.g.
    *   // public static String beanName = "<Bean Name>";
    *   // public static String getJavaScriptBeanName() { return beanName; }
    *   // public String getJavaScriptBeanName() { return beanName; }
    */
   public abstract String getJavaScriptBeanName();
   /**
    * overide this function to define all variables for this bean
    * 
    * @return
    */
   protected abstract Vector<String> getApplicationVariables();
   /**
    *  overide this function to define the variables types.  This corresponds
    *  with the collection of variables above.
    * @return
    */
   protected abstract Vector<Class> getApplicationVariableTypes();   

   /**
    * overide this function to define a cookieName to load cookie state from
    */
   public abstract String getCookieName();
   
   /**
    *  used to store the bean variables
    */
   private Map<String, Object> beanVariables = new HashMap<String,Object>();
   
   /**
    * Used internally to access the variables in this bean.
    * Each Bean should implement getters and setters that return this function like
    * return get("<BeanAttribute>");
    * 
    * @param attribute
    * @return
    */
   protected Object get(String attribute) {
      return beanVariables.get(attribute);      
   }
   
   /**
    * Used internally to access the variables in this bean, while
    * also providing access to our useful generic methods.
    * Each Bean should implement getters and setters that call this function like
    * set("<BeanAttribute>","<Value>");
    * 
    * @param attribute
    * @return
    */
   protected void set(String attribute, Object value) {
      beanVariables.put(attribute, value);
      isDirty = true;
   }

   /*
   public void persist(HttpServletRequest request, HttpServletResponse response) throws UnsupportedEncodingException {
      if (isDirty) {
         String queryString = "";
         for (int i = 0;i < getApplicationVariables().size();i++) {
            String key = getApplicationVariables().elementAt(i);
            Object value = get(key);
            if (i > 0) 
               queryString += '&'; 
            System.err.println(key + " = " + value);
            queryString += URLEncoder.encode(key,"UTF-8") + "=" + (value == null ? "" : URLEncoder.encode(value.toString(),"UTF-8"));
            
         }
         Cookie cookie = new Cookie(getCookieName(),queryString);
         cookie.setPath("/");
         response.addCookie(cookie);
         
         request.setAttribute(getCookieName(),this);
      }
   }
   */
   
   
   /**
    * Persist this Bean for the duration of the request.  This is a helper
    * for creating per request caching from a static loader 
    * @param request
    * @param response
    */
   protected void persistToRequest(HttpServletRequest request, HttpServletResponse response) {
      request.setAttribute(getCookieName(),this);   
   }
   
   protected static JavaScriptBean isPersistedInRequest(String cookieName, HttpServletRequest request, HttpServletResponse response)  {
      JavaScriptBean bean = (JavaScriptBean)request.getAttribute(cookieName);
      return bean;
   }
/*
   private void copyFrom(JavaScriptBean from) {
      for (int i = 0;i < getApplicationVariables().size();i++) {
         String key = getApplicationVariables().elementAt(i);
         Object value = from.get(key);
         set(key,value);
      }
   }
*/   
   /**
    * This method loads the bean variables.  We check to see if any changes have
    * been persisted, and return those if they exist.  Otherwise we load the
    * this bean from the cookie.  This allow us to implement reload,
    * Each bean should implement one constructor, and it should call this function
    * 
    *  // public <BeanName>(HttpServletRequest request, HttpServletResponse response) {
    *  //    load(request,response);
    *  //    
    *  //    // ... add any default settings here
    *  // }
    * 
    * @param request
    * @param response
    */
   protected void load(HttpServletRequest request, HttpServletResponse response) {

      /**
       * This seems like a less then perfect way to cache beans on a per request basic
       * When bean2 is created, values are copied from bean1, however changes to bean2
       * or bean1 won't be shared.  I think to get around this will need to share the 
       * same exact bean object somehow, or store their contents is a statically shared
       * location.
       * 
       * Fixed by Brian -
       * Create a static getInstance method like:
       * 
       */
      //JavaScriptBean bean = (JavaScriptBean) request.getAttribute(getCookieName());
      //if (bean != null) {
      //   copyFrom(bean);
      //   return;
      //}
      
      String cookieValue = CommonUtil.getCookieValue(request, getCookieName());
      if (cookieValue != null) {
         CommonUtil.setBackupQueryString(request, cookieValue);
      }
      
      for (int i = 0;i < getApplicationVariables().size();i++) {
         String key = getApplicationVariables().elementAt(i);
         
         String value = CommonUtil.getParameter(request, key);
         if (value != null) {
            Class type = getApplicationVariableTypes().elementAt(i);
            
//            log.debug("  KEY: " + key + " VALUE: " + value);
            
            /**
             * This will make the String "undefined" useless for real strings
             */
//            if (value == null || value.equals("undefined")) {
//               set(key,null);
            
            if (type == String.class) {
               set(key, value);
            } else if (type == Integer.class) {
               try {
                  Integer intValue = Integer.parseInt(value);
                  set(key, intValue);
               } catch (NumberFormatException nfe) {
                  log.warn("Parameter " + key + " is not of type Integer");
               }
            } else if (type == Long.class) {
               try {
                  Long longValue = Long.parseLong(value);
                  set(key, longValue);
               } catch (NumberFormatException nfe) {
                  log.warn("Parameter " + key + " is not of type Long");
               }
            } else if (type == Boolean.class) {
               Boolean booleanValue = Boolean.parseBoolean(value);
               set(key,booleanValue);
            }
            else {
               throw new RuntimeException("Invalid Paramter Type: " + type.getCanonicalName());
            }
         }
      }
   }
   
   
   
   
      
   /**
    * Generate the necassary object definitions to process a bean.  In order for a bean to function
    * correctly the bean must load the js object definition first.  Next the beanName
    * must be passed in the beans argument to the Web20Environment js object. Finally the initialize 
    * method of the Web20Environment js object is invoked and an instance of the object definition
    * is loaded into the Web20Evnirnment js object as beanName.
    * NOTE: like an other js function, this one outputs straight JavaScript and should be 
    * nested inside a <script type="text/javascript"> tag or JavaScript subrequest, or any other 
    * place straight JavaScript might be applicable.
    * 
    * @param out is a java.io.Writer we output to
    * @throws IOException when we have problems writing to out
    */
   public void jsDeclareObject(Writer out) throws IOException {
      out.write("function new_" + getJavaScriptBeanName() + "() {\n");
      out.write("  obj = {\n");
      for(int i = 0;i < getApplicationVariables().size();i++) {
         String var = getApplicationVariables().elementAt(i);
         Class type = getApplicationVariableTypes().elementAt(i);
         out.write("    " + var + ": ");
         if (type == String.class) {
            out.write(StringUtil.jsQuote((String)get(var)));
         } else if (type == Integer.class) {
            Integer value = (Integer)get(var);
            out.write((value == null ? "0" : value.toString()));
         } else if (type == Long.class) {
            Long value = (Long)get(var);
            out.write((value == null ? "0" : value.toString()));
         } else if (type == Boolean.class) {
            if ((Boolean)get(var)) {
               out.write("true");
            } else {
               out.write("false");
            }
         }
         out.write(",\n");
      }
      out.write("    /* The " + getJavaScriptBeanName() + " Bean */\n");
      out.write("    beanName: " + StringUtil.jsQuote(getJavaScriptBeanName())  + ",\n");
      out.write("    /* will set the specified bean value in js and persist */\n");
      out.write("    set: function(member, value) {\n");
      out.write("      eval('this.' + member + ' = value');\n");
      out.write("      this.persist();\n");
      out.write("    },\n");
      out.write("    /* persist this bean to it's cookie value */\n");
      out.write("    persist: function() {\n");
      out.write("      document.cookie = '" + getCookieName() + "=' + escape(this.serialize()) + ';path=/'\n"); 
      out.write("    },\n");
      out.write("    /* serialize this bean to a query string */\n");
      out.write("    serialize: function() {\n");
      out.write("      var state = ''\n");
      {
         for(int i = 0;i < getApplicationVariables().size();i++) {
            String var = getApplicationVariables().elementAt(i);
            Class type = getApplicationVariableTypes().elementAt(i);
            out.write("      if (this." + var + " != undefined) {\n");
            out.write("         if (state != '' ) state = state + '&';\n");
            out.write("         state = state +'" + var + "=' + escape(this." + var + ");\n"); 
            out.write("      }\n"); 
         }
      }
      out.write("      return state;\n");
      out.write("    },\n");
      out.write("    deserialize: function(state) {\n");      
      out.write("      var o = new Object;\n");
      out.write("      var parts = state.split('&')\n");
      out.write("      for (i = 0; i < parts.length;i++) {\n");
      out.write("        var part = parts[i];\n");
      out.write(Options.DEBUG("if (part == undefined) { w20.handleError('Unparseable state(1) ' + this.beanName + '=' + state); }\n"));
      out.write("        var kv = part.split('=');\n");
      out.write(Options.DEBUG("if (kv.length != 2) { w20.handleError('Unparseable state(2) ' + this.beanName + '=' + state); }\n"));
      out.write("        eval('o.' + kv[0] + '=unescape(kv[1]);');");
      out.write("      }\n");
      out.write("\n");
      {
         for(int i = 0;i < getApplicationVariables().size();i++) {
            String var = getApplicationVariables().elementAt(i);
            Class type = getApplicationVariableTypes().elementAt(i);
            out.write("      this." + var + " = o." + var + ";\n");
         }
      }
      out.write("      this.persist();\n");
      out.write("    },\n");
      
      out.write("    /* serialize and persist this bean from form elements */\n");
      out.write("    persistForm: function() {\n");
      {
         for(int i = 0;i < getApplicationVariables().size();i++) {
            String var = getApplicationVariables().elementAt(i);
            Class type = getApplicationVariableTypes().elementAt(i);
   
            out.write("      var " + var + "Value = w20.serializeElement(" + StringUtil.jsQuote(getAttributeId(var)) + ")\n"); 
            out.write("      if (" + var + "Value != undefined) { this." + var + " = " + var + "Value; }\n"); 
         }
         out.write("      this.persist()\n"); 
      }
      out.write("    }\n");
      
      out.write("  };\n");
      out.write("  return obj;\n");
      out.write("}\n");
               
   }
   

   /**
    * Outputs inline(single-line) JavaScript that will set and persist  the
    * specified attribute immediately, when the javascript is executed
    *
    *  
    * @param attribute
    * @param value
    * @return
    * @throws IOException 
    */
   public String jsGet(String attribute) throws IOException {
      return "w20.bean." + getJavaScriptBeanName() + ".get(" + StringUtil.jsQuote(attribute) +");";
   }

   
   /**
    * FIXME: Need to add type safety checks to jsSet
    *  
    * @param attribute
    * @param value
    * @return
    * @throws IOException 
    */
   public String jsSet(String attribute, String value) throws IOException {
      set(attribute,value);
      return "w20.bean." + getJavaScriptBeanName() + ".set(" + StringUtil.jsQuote(attribute) + "," + StringUtil.jsQuote(value) +");";
   }

   /**
    * Outputs inline(single-line) JavaScript that will set and persist  the
    * specified attribute immediately, when the javascript is executed
    *
    *  
    * @param attribute
    * @param value
    * @return
    * @throws IOException 
    */
   public String jsSet(String attribute, Integer value) throws IOException {
      set(attribute,value);
      return "w20.bean." + getJavaScriptBeanName() + ".set(" + StringUtil.jsQuote(attribute) + "," + StringUtil.jsQuote(value) +");";
   }

   
   /**
    * Outputs inline(single-line) JavaScript that will set and persist  the
    * specified attribute immediately, when the javascript is executed
    *
    *  
    * @param attribute
    * @param value
    * @return
    * @throws IOException 
    */
   public String jsSet(String attribute, Long value) throws IOException {
      set(attribute,value);
      return "w20.bean." + getJavaScriptBeanName() + ".set(" + StringUtil.jsQuote(attribute) + "," + StringUtil.jsQuote(value) +");";
   }

   /**
    * Outputs inline(single-line) JavaScript that will set and persist  the
    * specified attribute immediately, when the javascript is executed
    *
    *  
    * @param attribute
    * @param value
    * @return
    * @throws IOException 
    */
   public String jsSet(String attribute, Boolean value) throws IOException {
      set(attribute,value);
      return "w20.bean." + getJavaScriptBeanName() + ".set(" + StringUtil.jsQuote(attribute) + "," + StringUtil.jsQuote(value) +");";
   }

   /**
    * Load the html form elements by id for each attribute (see jsAttributeId) and persist changes.
    * This allows for w20 functionality to interact with forms in an easy and
    * natural fashion.  The following form elements are supported:<br>
    * <br>
    * *) text input 
    * *) hidden input
    * *) textarea
    * *) radio
    * *) checkbox
    * 
    * @param elementId
    * @return
    */
   public String jsPersistForm() {
      return "w20.bean." + getJavaScriptBeanName() + ".persistForm();\n";
   }
   
   /**
    * Works in conjunction with jsPersistForm.  Assign form elements to 
    * bean attributes be setting the html element id to this value.  Then
    * on onSubmit or any other JavaScript event persist the form using
    * jsPersistForm.  This will usually be followed be reloading or
    * forwarding a context. 
    * 
    * @param attribute
    * @return
    */
   public String getAttributeId(String attribute) {
      return getCookieName() + "__" + attribute;
   }

   
   /**
    * 
    * @param out
    * @throws IOException 
    */
   public String jsPersist() throws IOException {
      StringBuffer out = new StringBuffer();
      for(int i = 0;i < getApplicationVariables().size();i++) {
         String var = getApplicationVariables().elementAt(i);
         Class type = getApplicationVariableTypes().elementAt(i);
         out.append("w20.bean." + getJavaScriptBeanName() + "." + var + "=");
         if (type == String.class) {
            out.append(StringUtil.jsQuote((String)get(var)));
         } else if (type == Integer.class) {
            Integer value = (Integer)get(var);
            out.append(String.valueOf(value == null ? 0 : value));
         } else if (type == Long.class) {
            Long value = (Long)get(var);
            out.append(String.valueOf(value == null ? 0 : value));
         } else if (type == Boolean.class) {
            if ((Boolean)get(var)) {
               out.append("true");
            } else {
               out.append("false");
            }
         }
         out.append(";");
      }
      out.append("w20.bean." + getJavaScriptBeanName() + ".persist();");
      return out.toString();
   }


   
}
