package com.getsnappy.snappyj;

import java.io.IOException;
import java.io.Writer;
import java.util.Collection;
import java.util.Iterator;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import com.getsnappy.util.ReflectionCache.ClassCache;

/**
 * This class represents the core of our web 2.0 functionality
 * In order to use our library do the following:
 * <br>
 * <li>Define all contexts (contexts are reloadable areas on your page)</li>
 * <li>Define all beans (ALL server to browser communication is done through beans)</li>
 * <li>Instantiate this class into an object (call our constructor via new)</li>
 * <li>Call that objects jsInitialize within a <script type="text/javascript">
 * element.</li>
 * <li>Set default value's for bean's inside of onIntialLoad()</li>
 * <li>output the content of your page, which can contains both static
 * html and contexts included synchronously and/or asynchronously.</li>
 * <br>
 * 
 * @author brian
 *
 */
public class SnappyJEnvironment {

   String url;
   Collection<Context> contexts;
   Collection<JavaScriptBean> beans;
   String jsDoOnHistoryChange;

   /**
    * The only way to create an environment
    * @param url
    * @param contexts
    * @param beans
    * @param jsDoOnHistoryChange - This will be triggered when a client hits the back or forward button in their
    * browser right after the beans are restored to their corresponding state.  At least one context should be reloaded.
    */
   public SnappyJEnvironment(String url, Collection<Context> contexts, Collection<JavaScriptBean> beans, String jsDoOnHistoryChange) {
      this.url = url;
      this.contexts = contexts;
      this.beans = beans;
      this.jsDoOnHistoryChange = jsDoOnHistoryChange;
   }
   
   
   /**
    * Perform jsDo only if this is the initial page load.  This must be called
    * before calling setTitle with addToHistory set to true, or it will never
    * fire jsDo.  Also if you reload a page that has never called setTitle with
    * addToHisotry set to true, jsDo will be executed.
    * <br>
    * This is typically used to set default bean values for your master web page.
    * since bean state is persisted through cookies it's often necassary to 
    * 'reset' them when a user comes through the front door again.  If you reset
    * them globally then when a client reloads his browser, instead of being reloading
    * to his current state, the beans will be reset to their initial value.
    * <br>
    * 
    * Note: we distinguish
    * between initial loads by checking for the existence of a static js var
    * w20.isFirstTime.  Even if it is undefined this could be a page reload.
    * We check for the presence of an anchor tag used in history tracking.
    * So only if 
    * w20.isFirstTime == undefined and document.location.href.indexOf("#Page-") != -1
    * will jsDo execute
    * @param jsDo - javascript to execute
    * @throws IOException 
    */
   public static void jsIfInitialLoad(Writer out, String jsDo) throws IOException {
      out.write("if (w20.highHistoryPos == undefined && document.location.href.indexOf('#Page-') == -1) {\n");
      out.write(jsDo);
      out.write("}\n");
   }

   /**
    * 
    * @param out
    * @param jsDo
    * @throws IOException
    */
   public static void jsIfAsyncRequest(Writer out, String jsDo, HttpServletRequest request) throws IOException {
      if (request.getRequestURI().startsWith(Options.getServletForwarderUrl())) {
         out.write(jsDo);
      }
   }

   
   /**
    * I'm not actually sure what the applied use of this would be.  But it indicates that 
    * there was at least one bean snapshot added to history, and that the reload button
    * was hit.  It would be easy to create a popup that prompts the user as to whether
    * they want to start over or continue their w20 experience.  The function DOES NOT APPLY
    * to context reloads, only when the browser reload button is hit.
    *
    * @param jsDo - javascript to execute
    * @throws IOException 
    */
   public void jsIfReload(Writer out, String jsDo) throws IOException {
      out.write("if (w20.highHistoryPos == undefined && (document.location.href.indexOf('#Page-') != -1 && document.location.href.indexOf('#W20') != -1)) {\n");
      out.write(jsDo);
      out.write("}\n");
   }

   
   /**
    * Fired when this page is NOT the result of browser refresh.
    *
    * @param jsDo - javascript to execute
    * @throws IOException 
    */
   public static void jsIfNotReloadOrHistoryChange(Writer out, String jsDo, HttpServletRequest request) throws IOException {
      if (request.getParameter("W20IsHistoryChange") == null) {
         out.write("if ((w20.highHistoryPos == undefined && (document.location.href.indexOf('#Page-') == -1 && document.location.href.indexOf('#W20') == -1)) " +
      		          "|| (w20.highHistoryPos != undefined) " + 
      		          "||  (document.location.href.indexOf('#Page-') == -1)) {\n");
         out.write(jsDo);
         out.write("}\n");
      }
   }

   
   /**
    * Get a collection of all global beans.
    * 
    * @return
    */
   public Collection getBeans() {
      return beans;
   }

   
   /**
    * Get a collection of all global beans.
    * 
    * @return
    */
   public Collection getContexts() {
      return contexts;
   }

   /**
    * Get the url of the web page or servlet that is including and using this Web20Environment
    * 
    * @return
    */
   public String getUrl() {
      return url;
   }
   
   /**
    * This will create object declarations for each context and bean
    * within the web 2.0 envirnment.  This creates the base of the
    * libraries functionality, and should typically be called before any
    * other jsXXX call.  jsXXX calls indicate that straight javascript is
    * either returned (inline) our outputed to the out Writer param.
    * The jsInitialize would typically be called within a
    * <script type="text/javascript> node.
    * 
    * NOTE: This function also sets up the JavaScriptErrorHandler
    * defined as in Options, by default the DefaultJavaScriptErrorHandler.
    * @param request TODO
    * 
    * @throws IOException 
    * @throws ClassNotFoundException 
    * @throws IllegalAccessException 
    * @throws InstantiationException 
    */
   public void jsInitialize(Writer out) throws IOException, InstantiationException, IllegalAccessException, ClassNotFoundException {
      for (Iterator<Context> i = contexts.iterator();i.hasNext();) {
         Context context = i.next();
         context.jsDeclareObject(out);
      }
      for (Iterator<JavaScriptBean> i = beans.iterator();i.hasNext();) {
         JavaScriptBean bean = i.next();
         bean.jsDeclareObject(out);
      }
      
      
      out.write("w20.initialize('");
      out.write(url);
      out.write("','");
      out.write(CommaSeparatedContexts(contexts));
      out.write("','");
      out.write(CommaSeparatedBeans(beans));
      out.write("');\n");
      JavaScriptErrorHandler handler = (JavaScriptErrorHandler)(ClassCache.forName(Options.getErrorHandlerClass()).newInstance());
      handler.jsDeclareObject(out);
      
      out.write("w20.onHistoryChange = function() {\n");
      out.write(jsDoOnHistoryChange);
      out.write("}\n");

   }


   /**
    * For now this method is not public,
    * 
    * Call this method to output the JavaScript required to replace an element asynchronously.  IndicatorStart
    * and IndicatorFinish should resolve to javascript functions and can be null.  They are used to implement
    * JavaScriptLoadingIndicator's.  Callbacks should resolve to a javascript function or null as well.  Callbacks
    * are useful for a wide variety of things, and are execute write after the jIndicatorFinish, which in turn is
    * after the element was replaced successfully asynchronously.
    * 
    * @param id
    * @param className
    * @param callback - this must resolve a JavaScript function.  You will need to wrap with StringUtil.dontParse()
    */
   static String jsReplaceElementAsync(String id,String url,String callback, String indicatorStart, String indicatorFinish) {
      return "w20.replaceElementAsync(" + StringUtil.jsQuote(id) + "," + StringUtil.jsQuote(url) + "," + StringUtil.jsQuote(callback) + 
      "," + StringUtil.jsQuote(indicatorStart) + "," + StringUtil.jsQuote(indicatorFinish) + ");";
   }

   
   static String jsReplaceElementAsync(String id, String url,String callback) {
      return jsReplaceElementAsync(id, url, callback, null, null);
   }

   
   static String jsReplaceElementAsync(String id, String className, String queryString, 
         String callback, String indicatorStart, String indicatorFinish) {
      return jsReplaceElementAsync(id, Options.getServletForwarderUrl() + "?applicationClassName=" + className +
            ((queryString == null || queryString.equals("")) ? "" : "&") + queryString, callback,indicatorStart,indicatorFinish);
   }

   /**
    * A utility function for performing an asynchronous replacement of an element with the
    * contents of an application.  This allows for web 2.0 interactions that do not require
    * fancy loading indicators.  Because of the nature of web calls I would recommend against
    * using this function when serving to a global or distant audience because the delay between 
    * performing the action leading to the function call and the swapping of the div will make
    * for a less than desirable user experience.
    * 
    * 
    * @param id - the id attribute of the html element to be replaced (must exist in dom)
    * @param className - the className of an ApplicationInstance to execute
    * @param queryString - the query string to pass to the Application (should not begin with a ? or &)
    * @param callback - a javascript function (will be called like callback() if defined)
    * @return inline javascript
    */
   public static String jsReplaceElementAsync(String id, String className, String queryString, 
         String callback) {
      return jsReplaceElementAsync(id, className, queryString, callback, null, null);
   }

   
   /**
    * Call this method to output the JavaScript required to create 
    * make a popup window of the specified deminsions and 
    * id attributes, and LoadingIndicator sychronouly,
    * and then asynchronously The className with queryString.
    * 
    * @param src - this should resolve to a the element which is requesting this popup.  You will need to wrap with StringUtil.dontParse()
    * @param id - this is the id attribute that will be assigned to the web 2.0 popup(div) element
    * @param className - the className of the ApplicationInstance or HttpServlet that will be loaded asynchronously
    * @param callback - this must resolve a JavaScript function.  You will need to wrap with StringUtil.dontParse()
    */
   protected static String jsDisplayNonBlockingPopup(String src,String id,String className,String queryString, String callback) {
      return "w20.displayNonBlockingPopup(" + StringUtil.jsQuote(id) + "," + StringUtil.jsQuote(className) + "," + StringUtil.jsQuote(callback) + ");";
   }

   /**    
    * Setting the title and adding history are combined for a reason.
    * Typically you would design a document with the title at the top,
    * followed by the body.  But because we need the title to be the 
    * previous pages title at the time we take the bean snapshot and perform
    * the addToHistory magic, we do it backwords inside of Application's.
    * Only after all bean's are set to the correct value to reload the 
    * the current web page or servlet to it's current state, should you 
    * call setTitle with addToHistory.  Calling jsSetTitle with addToHistory
    * set to false and then later within an applications execution calling
    * jsSetTitle again with addToHistory set to true will also result in 
    * the first call's title param replacing the previous pages title in 
    * history.
    * 
    * @param title - may be null indicating no title change (quite common)
    * @param addToHistory - should the current state be saved so that a client
    * can use thier browsers built-in history to navigate back to it's initial
    * or previously saved state
    *
    * @return inline javascript
    * @throws IOException 
    */
   public static void jsSetTitle(Writer out, String title) throws IOException {
      out.write(Options.DEBUG("if (w20 == undefined) { w20.handleError('w20 is not loaded'); }"));
      out.write("w20.setTitle(" + StringUtil.jsQuote(title) + ");\n");
      
   }
   
   public static void jsAddToHistory(Writer out, HttpServletRequest request) throws IOException {
      out.write(Options.DEBUG("if (w20 == undefined) { w20.handleError('w20 is not loaded'); }"));
      
      // don't add to history
      if (request.getParameter("W20IsHistoryChange") == null) {
         out.write("w20.addToHistory();\n");
      }
      
   }


   /**
    * Convenience function
    * 
    * @param c
    * @return
    */
   private String CommaSeparatedBeans(Collection<JavaScriptBean> c) {
      StringBuffer sb = new StringBuffer();
      for (Iterator<JavaScriptBean> i = c.iterator();i.hasNext();) {
         if (sb.length() > 0) sb.append(","); 
         
         sb.append(i.next().getJavaScriptBeanName());
      }
      
      return sb.toString();
   }

   /**
    * Convenience function
    * 
    * @param c
    * @return
    */
   private String CommaSeparatedContexts(Collection<Context> c) {
      StringBuffer sb = new StringBuffer();
      for (Iterator<Context> i = c.iterator();i.hasNext();) {
         if (sb.length() > 0) { sb.append(","); }
         
         sb.append(i.next().getContextId());
      }
      
      return sb.toString();
   }


}
